########################################################################
# File name: lmc.py
# This file is part of: aioxmpp
#
# LICENSE
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program.  If not, see
# <http://www.gnu.org/licenses/>.
#
########################################################################
import aioxmpp.xso as xso

from aioxmpp.utils import namespaces

from ..stanza import Message


namespaces.xep0308_replace = "urn:xmpp:message-correct:0"


class Replace(xso.XSO):
    """
    A marker indicating that the stanza is the correction of another one

    .. attribute:: id_

       The identifier of the stanza to correct.
    """

    TAG = namespaces.xep0308_replace, "replace"

    id_ = xso.Attr(
        "id",
    )


Message.xep0308_replace = xso.Child([Replace])
